# coding: utf-8
# create by tongshiwei on 2019-8-27

__all__ = ["Splitter", "split", "divide"]

from ..indicesor import Indicesor


class Splitter(object):
    def __init__(self, key_iterable, shuffle_indices=False, *args, silent=False, **kwargs):
        self.indices = Indicesor.get_indices(key_iterable, shuffle=shuffle_indices, silent=silent)
        self.silent = silent

    def _split(self, source, target):
        raise NotImplementedError

    def split(self, *source_target_iterable):
        for source, target in source_target_iterable:
            self._split(source, target)

    def __call__(self, *source_target_iterable):
        return self.split(*source_target_iterable)


def split(splitter, *source_target_iterable):
    splitter(*source_target_iterable)


def divide(splitter_class, *source_target_iterable, **splitter_params):
    splitter = splitter_class(**splitter_params)
    split(splitter, *source_target_iterable)



# coding: utf-8
# create by tongshiwei on 2019-8-27

__all__ = ["Indicesor"]

import io
import random
from copy import deepcopy
from tqdm import tqdm
from longling import config_logging

logger = config_logging(logger="longling")


class Indicesor(object):
    def __init__(self, key_iterable=None, shuffle=False, *args, silent=False, **kwargs):
        self.silent = silent
        if key_iterable is None:
            self.indices = None
        else:
            self.indices = self.initial_indices(key_iterable)
            if shuffle is True:
                random.shuffle(self.indices)

    def __call__(self, key_iterable=None, *args, **kwargs):
        if key_iterable is None:
            assert self.indices is not None
            return self.indices
        else:
            indices = self.initial_indices(key_iterable)
            random.shuffle(indices)
            return indices

    def initial_indices(self, key_iterable):
        if hasattr(key_iterable, "__len__"):
            indices = range(len(key_iterable))
        else:
            def get_indices(_key_iterable):
                _indices = None
                for idx, _ in tqdm(enumerate(_key_iterable), "initial indices", disable=self.silent):
                    _indices = idx
                assert _indices is not None
                return list(range(_indices))

            if isinstance(key_iterable, io.TextIOWrapper):
                offset = key_iterable.tell()
                indices = get_indices(key_iterable)
                key_iterable.seek(offset)
            else:
                try:
                    key_iterable = deepcopy(key_iterable)
                except TypeError as e:
                    logger.warning("Skip deepcopy procedure")
                    logger.warning(e)
                indices = get_indices(key_iterable)
        self.indices = indices
        return self.indices

    @staticmethod
    def get_indices(key_iterable, shuffle=False, silent=False):
        return Indicesor(key_iterable, shuffle=shuffle, silent=silent).indices



# coding: utf-8
# 2020/1/10 @ tongshiwei

import pathlib
import warnings
from heapq import nlargest
from longling.ML.toolkit.analyser import get_max, get_by_key, key_parser
from longling import Configuration, path_append
from longling import dict2pv, list2dict

import json
import sqlite3
import os


def _key(x):
    try:
        return float(x)
    except ValueError:
        return float(json.loads(x)["default"])


def show(key, exp_id=None, res_dir="./", nni_dir=path_append(os.environ.get("HOME", "./"), "nni/experiments"),
         only_final=False,
         with_keys=None, with_all=False):
    """
    cli alias: ``nni show``

    Parameters
    ----------
    key
    exp_id
    res_dir
    nni_dir
    only_final
    with_keys
    with_all

    Returns
    -------

    """
    if exp_id is None:
        exp_id = pathlib.PurePath(os.path.abspath(res_dir)).name
    nni_dir = path_append(nni_dir, exp_id)
    sqlite_db = path_append(nni_dir, "db", "nni.sqlite", to_str=True)
    print(sqlite_db)
    conn = sqlite3.connect(sqlite_db)
    c = conn.cursor()
    if only_final:
        cursor = c.execute("SELECT trialJobId FROM MetricData WHERE type='FINAL';")
    else:
        cursor = c.execute("SELECT DISTINCT trialJobId FROM MetricData;")
    trial_dir = path_append(nni_dir, "trials")
    result = []
    for trial in [row[0] for row in cursor]:
        with open(path_append(trial_dir, trial, "parameter.cfg")) as f:
            trial_params = json.load(f)["parameters"]
        trial_res = path_append(res_dir, trial, "result.json", to_str=True)
        value, appendix = get_max(trial_res, key, with_keys=with_keys, with_all=with_all)
        if with_keys is not None or with_all is True:
            result.append([trial, trial_params, value, dict(appendix[key])])
        else:
            result.append([trial, trial_params, value])
    conn.close()
    result.sort(key=lambda x: float(x[2][key]), reverse=True)
    return result


def show_top_k(k, exp_id=None, exp_dir=path_append(os.environ.get("HOME", "./"), "nni/experiments")):
    """
    cli alias: ``nni k-best``

    Parameters
    ----------
    k
    exp_id
    exp_dir

    Returns
    -------

    """
    if exp_id:
        exp_dir = path_append(exp_dir, exp_id)
    sqlite_db = path_append(exp_dir, "db", "nni.sqlite", to_str=True)
    print(sqlite_db)
    conn = sqlite3.connect(sqlite_db)
    c = conn.cursor()
    cursor = c.execute("SELECT trialJobId, data FROM MetricData WHERE type='FINAL';")
    _ret = []
    top_k = nlargest(k, [row for row in cursor], key=lambda x: _key(x[1]))
    trial_dir = path_append(exp_dir, "trials")
    for trial, result in sorted(top_k, key=lambda x: _key(x[1]), reverse=True):
        with open(path_append(trial_dir, trial, "parameter.cfg")) as f:
            trial_params = json.load(f)["parameters"]
            _ret.append([trial, result, trial_params])
    conn.close()

    return _ret


class BaseReporter(object):
    def intermediate(self, data):
        raise NotImplementedError

    def final(self):
        raise NotImplementedError


def get_params(received_params: dict, cfg_cls: Configuration):
    cfg_params = {}
    u_params = {}

    path, _ = dict2pv(cfg_cls.vars())

    keys = {p[-1]: p for p in path}

    for k, v in received_params.items():
        if k in cfg_cls.vars():
            cfg_params[k] = v
        else:
            if k in keys:
                cfg_params.update(list2dict(keys[k], v))
            else:
                u_params[k] = v

    return cfg_params, u_params


def prepare_hyper_search(cfg_kwargs: dict, cfg_cls, reporthook=None, final_reporthook=None,
                         primary_key=None, reporter_cls=None, with_keys: (list, str, None) = None, dump=False):
    try:
        import nni
        from nni import get_next_parameter, report_intermediate_result, report_final_result

        assert primary_key is not None

        if isinstance(with_keys, str):
            if ";" in with_keys:
                with_keys = with_keys.split(";")
            else:
                with_keys = [with_keys]
        elif with_keys is None:
            with_keys = []

        class Reporter(BaseReporter):
            def __init__(self):
                self.datas = []

            def intermediate(self, data):
                feed_dict = {'default': float(data[primary_key])}
                for key in with_keys:
                    feed_dict[key] = get_by_key(data, key_parser(key))
                report_intermediate_result(feed_dict)
                self.datas.append(data)

            def final(self):
                final_result = get_max(self.datas, primary_key, with_keys=";".join(with_keys) if with_keys else None)
                feed_dict = {
                    'default': float(final_result[0][primary_key])
                }
                appendix_dict = dict(final_result[1][primary_key])
                for key in with_keys:
                    feed_dict[key] = appendix_dict[key]
                report_final_result(feed_dict)

        rc = Reporter() if reporter_cls is None else reporter_cls
        reporthook = reporthook if reporthook is not None else rc.intermediate
        final_reporthook = final_reporthook if final_reporthook is not None else rc.final
        cfg_cls_params, hyper_params = get_params(get_next_parameter(), cfg_cls)
        using_nni_tag = True if cfg_cls_params or hyper_params else False
        cfg_kwargs.update(cfg_cls_params)
        cfg_kwargs["hyper_params"].update(hyper_params)
        if using_nni_tag is True and dump is True:
            cfg_kwargs["workspace"] = cfg_kwargs.get("workspace", "") + path_append(
                nni.get_experiment_id(), nni.get_trial_id(), to_str=True
            )
        return cfg_kwargs, reporthook, final_reporthook, dump

    except ModuleNotFoundError:
        warnings.warn("nni package not found, skip")
        return cfg_kwargs, reporthook, final_reporthook, dump



# coding: utf-8
# create by tongshiwei on 2018/7/14

from .ValueMonitor import ValueMonitor, MovingValue

__all__ = ["LossMonitor", "MovingLoss"]


class LossMonitor(ValueMonitor):
    @property
    def losses(self):
        return self._values


class MovingLoss(MovingValue, LossMonitor):
    pass



# coding: utf-8
# 2019/12/20 @ tongshiwei

import json
from longling.lib.candylib import as_list
from longling import PATH_TYPE, loading

__all__ = ["get_max", "key_parser", "get_by_key"]


def key_parser(key):
    if ":" in key:
        # prf:0:f1
        return key.split(":")
    return key


def get_by_key(data, parsed_key):
    _key = key_parser(parsed_key)
    _data = data
    if isinstance(_key, list):
        for k in _key:
            _data = _data[k]
    else:
        _data = _data[_key]
    return _data


def get_max(src: (PATH_TYPE, list), *keys, with_keys: (str, None) = None, with_all=False):
    keys = as_list(keys)

    with_keys = [] if with_keys is None else with_keys.split(";")

    result = {
        key: None for key in keys
    }

    result_appendix = {
        key: None for key in keys
    }

    for data in loading(src):
        for key in result:
            _data = get_by_key(data, parsed_key=key_parser(key))
            if result[key] is None or _data > result[key]:
                result[key] = _data
                if with_all:
                    result_appendix[key] = data
                elif with_keys:
                    result_appendix[key] = [(_key, get_by_key(data, key_parser(_key))) for _key in with_keys]

    return result, result_appendix



# coding: utf-8
# 2019/10/12 @ tongshiwei

import json
from longling.lib.candylib import as_list
from .select import get_max

__all__ = ["select_max", "arg_select_max"]


def key_parser(key):
    if ":" in key:
        # prf:0:f1
        return key.split(":")
    return key


def get_by_key(data, parsed_key):
    _key = key_parser(parsed_key)
    _data = data
    if isinstance(_key, list):
        for k in _key:
            _data = _data[k]
    else:
        _data = _data[_key]
    return _data


def _select_max(src, *keys, with_keys: (str, None) = None, with_all=False):
    result, result_appendix = get_max(src, *keys, with_keys=with_keys, with_all=with_all)

    for item in result.items():
        print(item, end='')
        if result_appendix[item[0]]:
            print(" ", result_appendix[item[0]])
        else:
            print()


def arg_select_max(*keys, src, with_keys=None, with_all=False):
    """
    cli alias: ``amax``

    Parameters
    ----------
    keys
    src
    with_keys
    with_all

    Returns
    -------

    """
    _select_max(src, *keys, with_keys=with_keys, with_all=with_all)


def select_max(src, *keys, with_keys=None, with_all=False):
    """
    cli alias: ``max``

    Parameters
    ----------
    src
    keys
    with_keys
    with_all

    Returns
    -------

    """
    _select_max(src, *keys, with_keys=with_keys, with_all=with_all)



# coding: utf-8
# 2019/9/21 @ tongshiwei


def legal_input(__promt: str,
                __legal_input: set = None, __illegal_input: set = None, is_legal=None,
                __default_value: str = None):  # pragma: no cover
    """
    To make sure the input legal, if the input is illegal, the user will be asked to retype the input.

    Input being illegal means the input either not in __legal_input or in __illegal_input.

    For the case that user do not type in anything, if the __default_value is set, the __default_value will be used.
    Else, the input will be treated as illegal.

    Parameters
    ----------
    __promt: str
        tips to be displayed
    __legal_input: set
        the input should be in the __legal_input set
    __illegal_input: set
        the input should be not in the __illegal_input set
    is_legal: function
        the function used to judge whether the input is legal, by default, we use the inner __is_legal function
    __default_value: str
        default value when user type in nothing

    """
    inp = input(__promt)
    __legal_input = __legal_input if __legal_input else set()
    __illegal_input = __illegal_input if __illegal_input else set()

    def __is_legal(__inp):
        if (not __legal_input or __inp in __legal_input) and __inp not in __illegal_input:
            # whether the input either not in __legal_input or in __illegal_input
            return True
        elif not inp and __default_value:
            # If the __default_value is set, the __default_value will be used.
            # Else, the input will be treated as illegal.
            return True
        return False

    is_legal = is_legal if is_legal else __is_legal

    while not is_legal(inp):
        # the user will be asked to retype the input.
        inp = input(__promt)
    if not inp and __default_value:
        return __default_value
    return inp


def default_legal_input(tip, __legal_input: set = None, __illegal_input: set = None, is_legal=None,
                        __default_value: str = None, cursor=">"):
    __promt = tip

    legal_tip = []

    if __legal_input is not None:
        legal_tip.append("/".join(__legal_input))

    if __default_value is not None:
        legal_tip.append("default is %s" % __default_value)

    if legal_tip:
        __promt = tip + " (%s)" % ", ".join(legal_tip)

    __promt += " %s " % cursor
    return legal_input(__promt, __legal_input=__legal_input, __illegal_input=__illegal_input, is_legal=is_legal,
                       __default_value=__default_value)


def binary_legal_input(tip, _y="y", _n="n", _default="y"):
    assert _default in [_y, _n, None]
    return True if legal_input("%s (%s/%s, default is %s) < " % (tip, _y, _n, _default),
                               __default_value=_default) == _y else False
